# Query responses
EXPECTED_GET_DATA_EXPORT_URL = "getDataExportUrl"

# Data Export Names
MONITORS_DATA_EXPORT_NAME = "MONITORS"
ASSETS_DATA_EXPORT_NAME = "ASSETS"
ALERTS_DATA_EXPORT_NAME = "ALERTS"
EVENTS_DATA_EXPORT_NAME = "EVENTS"

# Supported Schemes
FILE_SCHEME = "file"
S3_SCHEME = "s3"
SCHEME_DELIM = "://"

# Data Export headers
LIST_DATA_EXPORTS_HEADERS = ["Data Export (Name)", "Description"]
